!****************************************************************************
!****************************************************************************
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
!****************************************************************************
!****************************************************************************


!****************************************************************************
!****************************************************************************
! Codecs. Mpeg video codec.
!
! (c)2001 Guillaume & Mathieu Legris.
!****************************************************************************
!****************************************************************************

.text


!****************************************************************************
! Class MpegVideoCodec.
!****************************************************************************
!----------------------------------------------------------------------------
! Edges.
!----------------------------------------------------------------------------
.align	5
.global _vu_codecs_mpegvideocodec_makeEdges
_vu_codecs_mpegvideocodec_makeEdges:
	mov.l		r8,@-r15
	mov.l		r9,@-r15

	mov.l		@r15+,r9
	mov.l		@r15+,r8
	rts
	 nop

!----------------------------------------------------------------------------
! Reconstruction.
!----------------------------------------------------------------------------
!struct ReconstructionInfo {
!	uint8 *frameReference, *frameForward;
!	int32 width, height;
!	int32 x, y;
!	int32 dx, dy;
!	int32 stride;
!	int32 roundingType;
!};

.align	5
! void reconstructBlock(ReconstructionInfo *ri, int32 isMacroBlock);
.global _vu_codecs_mpegvideocodec_reconstructBlock
_vu_codecs_mpegvideocodec_reconstructBlock:
	mov.l		r8,@-r15
	tst		r5,r5
	mov.l		r9,@-r15
	mov		#-8,r8 ! r8=-blockWidth=-8.
	mov.l		r10,@-r15
	bt			.notMacroBlock
	mov		#-16,r8 ! r8=-blockWidth=-16.
.notMacroBlock:

	mov.l		@(4*6,r4),r0 ! r0=dx.
	mov		#0,r9
	mov.l		@(4*7,r4),r1 ! r1=dy.
!	nop
	mov.l		@r4,r5 ! r5=frameReference.
	shar		r0 ! r0=dx>>1.
	mov.l		@(4*8,r4),r6 ! r6=stride.
	rotcl		r9
	mov.l		@(4*5,r4),r3 ! r3=y.
	shar		r1 ! r1=dy>>1.
	mov.l		@(4*4,r4),r2 ! r2=x.
	rotcl		r9 ! r9=copy function.

	mul.l		r6,r3 ! macl=y*stride.

	add		r2,r5 ! r5=frameReference+x.
!	nop
	add		r0,r2 ! r2=xs=x+(dx>>1).
	mov.l		@(4*2,r4),r0 ! r0=width.
	add		r1,r3 ! r3=ys=y+(dy>>1).
	mov.l		@(4*3,r4),r1 ! r1=height.

	sts		macl,r7 ! r7=y*stride.

	! Clipping.
	cmp/ge	r8,r3 ! if (ys<-blockWidth).
	bt			rb_skipy0
	mov		r8,r3
rb_skipy0:
	cmp/gt	r1,r3 ! if (ys>height).
	bf			rb_skipy1
	mov		r1,r3
rb_skipy1:

	mul.l		r6,r3 ! macl=ys*stride.

	cmp/ge	r8,r2 ! if (xs<-blockWidth).
	bt			rb_skipx0
	mov		r8,r2
rb_skipx0:
	neg		r8,r8 ! r8=blockWidth.

	sts		macl,r3 ! r3=ys*width;

	add		r7,r5 ! r5=frameReference+y*stride+x;
	mov.l		@(4*9,r4),r7 ! r7=roundingType.
	shll2		r9
	mov.l		@(4*1,r4),r4 ! r4=frameForward.

	cmp/gt	r0,r2 ! if (xs>width).
	bf			rb_skipx1
	mov		r0,r2
rb_skipx1:
	mova		blockCopyFunctions,r0

	mov.l		@(r0,r9),r0 ! r0=pointer on the copy function.
	add		r2,r4 ! r4=frameForward+xs.
	pref		@r5
	add		r3,r4 ! r4=frameForward+ys*stride+xs.

	mov		#1,r1
	pref		@r4
	and		r1,r7
!	nop
	xor		r1,r7
!	nop

	jmp		@r0
	 mov		r7,r9


	.align	5
blockCopyFunctions:
	.long		copyBlock, copyBlockV, copyBlockH, copyBlockHV
	.align	5


	!
	! No interpolation.
	!

copyBlock:
	mov		r8,r3
	shlr2		r3
	mov		r8,r7
	shlr		r3 ! r3=blockWidth/8.
	sub		r7,r6 ! r6=stride-blockWidth.

.c_loopY:
	pref		@r5
	mov		r3,r8

.c_loopX:
	mov.b		@r4+,r0
!	nop
	mov.b		@r4+,r1
!	nop
	mov.b		r0,@r5
	add		#1,r5
	mov.b		r1,@r5
	add		#1,r5

	mov.b		@r4+,r0
!	nop
	mov.b		@r4+,r1
!	nop
	mov.b		r0,@r5
	add		#1,r5
	mov.b		r1,@r5
	add		#1,r5

	mov.b		@r4+,r0
!	nop
	mov.b		@r4+,r1
!	nop
	mov.b		r0,@r5
	add		#1,r5
	mov.b		r1,@r5
	add		#1,r5

	mov.b		@r4+,r0
!	nop
	mov.b		@r4+,r1
	dt			r8
	mov.b		r0,@r5
	add		#1,r5
	mov.b		r1,@r5

	bf/s		.c_loopX
	 add		#1,r5

	pref		@r4
	dt			r7
	pref		@r5
	add		r6,r4
	bf/s		.c_loopY
	 add		r6,r5

	mov.l		@r15+,r10
!	nop
	mov.l		@r15+,r9
!	nop
	rts
	 mov.l	@r15+,r8


	!
	! Horizontal interpolation.
	!

copyBlockH:
	mov		r8,r3
	shlr2		r3
	mov		r8,r7
	shlr		r3 ! r3=blockWidth/8.
	sub		r7,r6 ! r6=stride-blockWidth.

.ch_loopY:
	pref		@r5
	mov		r3,r8

.ch_loopX:
	mov.b		@r4+,r0
!	nop
	mov.b		@r4+,r1
!	nop
	mov.b		@r4+,r2
	extu.b	r0,r0

	extu.b	r1,r1
!	nop
	add		r1,r0
!	nop
	add		r9,r0
!	nop
	shlr		r0
!	nop

	mov.b		r0,@r5
	extu.b	r2,r2
	mov		r1,r0
	add		r2,r0
	mov.b		@r4+,r1
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(1,r5)
	extu.b	r1,r1
	mov		r2,r0
	add		r1,r0
	mov.b		@r4+,r2
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(2,r5)
	extu.b	r2,r2
	mov		r1,r0
	add		r2,r0
	mov.b		@r4+,r1
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(3,r5)
	extu.b	r1,r1
	mov		r2,r0
	add		r1,r0
	mov.b		@r4+,r2
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(4,r5)
	extu.b	r2,r2
	mov		r1,r0
	add		r2,r0
	mov.b		@r4+,r1
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(5,r5)
	extu.b	r1,r1
	mov		r2,r0
	add		r1,r0
	mov.b		@r4,r2
	add		r9,r0
	shlr		r0
!	nop

	mov.b		r0,@(6,r5)
	extu.b	r2,r2
	mov		r1,r0
	add		r2,r0
!	nop
	add		r9,r0
	shlr		r0
!	nop

	dt			r8
	mov.b		r0,@(7,r5)
	bf/s		.ch_loopX
	 add		#8,r5

	pref		@r4
	dt			r7
	pref		@r5
	add		r6,r4
	bf/s		.ch_loopY
	 add		r6,r5

	mov.l		@r15+,r10
!	nop
	mov.l		@r15+,r9
!	nop
	rts
	 mov.l	@r15+,r8


	!
	! Vertical interpolation.
	!

copyBlockV:
	mov		r4,r3
	add		r6,r3
	sub		r8,r6 ! r6=stride-blockWidth.
	mov		r8,r10
	shlr2		r10
	mov		r8,r7
	shlr		r10 ! r10=blockWidth/8.

.cv_loopY:
	pref		@r5
	mov		r10,r8

.cv_loopX:
	mov.b		@r4+,r2
!	nop
	mov.b		@r3+,r1
!	nop

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@r5

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(1,r5)

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(2,r5)

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(3,r5)

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(4,r5)

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(5,r5)

	extu.b	r2,r0
	mov.b		@r4+,r2
	extu.b	r1,r1
!	nop
	add		r1,r0
	mov.b		@r3+,r1
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(6,r5)

	extu.b	r2,r0
!	nop
	extu.b	r1,r1
!	nop
	add		r1,r0
!	nop
	add		r9,r0
!	nop
	shlr		r0
	mov.b		r0,@(7,r5)

	dt			r8
	bf/s		.cv_loopX
	 add		#8,r5

	pref		@r3
	dt			r7
	pref		@r4
	add		r6,r3
	pref		@r5
	add		r6,r4
	bf/s		.cv_loopY
	 add		r6,r5

	mov.l		@r15+,r10
!	nop
	mov.l		@r15+,r9
!	nop
	rts
	 mov.l	@r15+,r8


	!
	! Horizontal and vertical interpolation.
	!

copyBlockHV:
	mov		r8,r10
	shlr2		r10
	mov.l		r11,@-r15
	shlr		r10 ! r10=blockWidth/8.
	mov		r4,r11
	add		r6,r11 ! r11=srcAddress+stride.
	mov		r8,r7 ! r7=blockWidth.
	sub		r8,r6 ! r6=stride-blockWidth.
	add		#1,r9

.chv_loopY:
	pref		@r5
	mov		r10,r8

.chv_loopX:
	mov.b		@r4+,r0
!	nop
	mov.b		@r11+,r1
!	nop
	mov.b		@r4+,r2
	extu.b	r0,r0
	mov.b		@r11+,r3
	extu.b	r1,r1
!	nop
	extu.b	r2,r2
!	nop
	extu.b	r3,r3

	add		r1,r0
	mov.b		@r4+,r1
	add		r3,r2
	mov.b		@r11+,r3
	add		r2,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r1,r1
!	nop
	extu.b	r3,r3
	mov.b		r0,@r5

	mov		r2,r0
	mov.b		@r4+,r2
	add		r3,r1
	mov.b		@r11+,r3
	add		r1,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r2,r2
!	nop
	extu.b	r3,r3
	mov.b		r0,@(1,r5)

	mov		r1,r0
	mov.b		@r4+,r1
	add		r3,r2
	mov.b		@r11+,r3
	add		r2,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r1,r1
!	nop
	extu.b	r3,r3
	mov.b		r0,@(2,r5)

	mov		r2,r0
	mov.b		@r4+,r2
	add		r3,r1
	mov.b		@r11+,r3
	add		r1,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r2,r2
!	nop
	extu.b	r3,r3
	mov.b		r0,@(3,r5)

	mov		r1,r0
	mov.b		@r4+,r1
	add		r3,r2
	mov.b		@r11+,r3
	add		r2,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r1,r1
!	nop
	extu.b	r3,r3
	mov.b		r0,@(4,r5)

	mov		r2,r0
	mov.b		@r4+,r2
	add		r3,r1
	mov.b		@r11+,r3
	add		r1,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r2,r2
!	nop
	extu.b	r3,r3
	mov.b		r0,@(5,r5)

	mov		r1,r0
	mov.b		@r4,r1
	add		r3,r2
	mov.b		@r11,r3
	add		r2,r0
!	nop
	add		r9,r0
!	nop
	shlr2		r0
!	nop
	extu.b	r1,r1
!	nop
	extu.b	r3,r3
	mov.b		r0,@(6,r5)

	mov		r2,r0
	add		r3,r1
!	nop
	add		r1,r0
!	nop
	add		r9,r0
	shlr2		r0
	dt			r8
	mov.b		r0,@(7,r5)

	bf/s		.chv_loopX
	 add		#8,r5

	pref		@r4
	dt			r7
	pref		@r5
	add		r6,r4
	pref		@r11
	add		r6,r5
	bf/s		.chv_loopY
	 add		r6,r11

	mov.l		@r15+,r11
!	nop
	mov.l		@r15+,r10
!	nop
	mov.l		@r15+,r9
!	nop
	rts
	 mov.l	@r15+,r8

!----------------------------------------------------------------------------
! Dequantization.
!----------------------------------------------------------------------------
.align	5
! Not optimized.
! void dequantizationMpeg1(int32 *block, int32 quantizer, uint8 *matrix);
.global _vu_codecs_mpegvideocodec_dequantizationMpeg1
_vu_codecs_mpegvideocodec_dequantizationMpeg1:
	pref		@r4
	add		#4,r4
	add		#1,r6
	mov		#3,r2

	mov		#63,r7
.dqmpeg1_loop:
	mov.l		@r4,r2
	mov.b		@r6+,r3
	mul.l		r5,r2
	sts		macl,r0
	mul.l		r3,r0
	sts		macl,r0
	shad		r2,r0
	dt			r7
	mov.l		r0,@r4
	bf/s		.dqmpeg1_loop
	 add		#4,r4

	rts
	 nop

.align	5
! void dequantizationH263(int32 *block, int32 quantizer);
.global _vu_codecs_mpegvideocodec_dequantizationH263
_vu_codecs_mpegvideocodec_dequantizationH263:
	pref		@r4
	add		#4,r4
	mov		#1,r0
	mov		r5,r6 ! r6=quantizer.
	sub		r0,r6 ! r6=quantizer-1.
!	nop
	or			r0,r6 ! r6=(quantizer-1)|1.
!	nop
	add		r5,r5 ! r5=quantizer<<1.
!	nop

	mov		#63,r7
.dqh263_loop:
	mov.l		@r4,r0
	mov		r6,r2

!	nop nop

	tst		r0,r0
	bt/s		.dqh263_zero
	 cmp/pz	r0
	bt			.dqh263_skip
	neg		r2,r2
.dqh263_skip:

	mul.l		r5,r0

!	nop nop
!	nop nop

	sts		macl,r1

!	nop nop
!	nop nop

	add		r1,r2
	mov.l		r2,@r4

.dqh263_zero:
	dt			r7
	bf/s		.dqh263_loop:
	 add		#4,r4

	rts
	 nop

!----------------------------------------------------------------------------
! Block.
!----------------------------------------------------------------------------
.align	5
! void addBlockIntra(int32 *block, uint8 *frame, int32 stride);
.global _vu_codecs_mpegvideocodec_addBlockIntra
_vu_codecs_mpegvideocodec_addBlockIntra:
	mov		#0,r2
	pref		@r5
	mov		#255,r3
!	nop
	extu.b	r3,r3 ! r3=255.
!	nop
	add		#-8,r6 ! r6=stride-8.
!	nop
	mov		#8,r1 ! r1=y loop count.

.loopY0:
	mov		#8,r7 ! r7=x loop count.

.loopX0:
	mov.l		@r4,r0
!	nop
	mov.l		r2,@r4
	add		#4,r4

	cmp/pz	r0
	bt			.skip00
	mov		r2,r0
.skip00:
	cmp/gt	r3,r0
	bf			.skip01
	mov		r3,r0
.skip01:

	dt			r7
	mov.b		r0,@r5

	bf/s		.loopX0
	 add		#1,r5

	pref		@r5
	dt			r1
	bf/s		.loopY0
	 add		r6,r5

	rts
	 nop

! void addBlockInter(int32 *block, uint8 *frame, int32 stride);
.global _vu_codecs_mpegvideocodec_addBlockInter
_vu_codecs_mpegvideocodec_addBlockInter:
	mov		#0,r2
	mov.l		r8,@-r15
	mov		#255,r3
!	nop
	extu.b	r3,r3 ! r3=255.
!	nop
	add		#-8,r6 ! r6=stride-8.
!	nop
	mov		#8,r8 ! r8=y loop count.
!	nop


.loopY1:
	mov		#8,r7 ! r7=x loop count.

.loopX1:
	mov.b		@r5,r1
!	nop
	mov.l		@r4,r0
!	nop
	mov.l		r2,@r4
	extu.b	r1,r1
	add		r1,r0
!	nop
	add		#4,r4

	cmp/pz	r0
	bt			.skip20
	mov		r2,r0
.skip20:
	cmp/gt	r3,r0
	bf			.skip21
	mov		r3,r0
.skip21:

	dt			r7
	mov.b		r0,@r5

	bf/s		.loopX1
	 add		#1,r5

	pref		@r5
	dt			r8
	bf/s		.loopY1
	 add		r6,r5

	rts
	 mov.l	@r15+,r8
!****************************************************************************
